import asyncio
import datetime as dt
from typing import Optional

import pandas as pd
from pydantic.dataclasses import dataclass
from stpmex.exc import NoOrdenesEncontradas
from stpmex.resources import Orden as OrdenSTPMex
from stpmex.types import Estado, TipoOperacion

from ..http import session
from ..utils import strftime, strptime
from .base import Resource

STP_BANK_CODE = 90646


@dataclass
class Orden(Resource, OrdenSTPMex):
    @classmethod
    async def _consulta_fecha(  # type: ignore
        cls, tipo: TipoOperacion, fechaOperacion: Optional[dt.date] = None
    ) -> pd.DataFrame:
        """
        Exclude the fechaOperacion if looking up transactions from the same
        day or when the fechaOperacion is in the future, in the event of this
        function being called during non-banking hours (9am – 6pm) / days.
        """
        endpoint = cls._endpoint + '/consOrdenesFech'
        consulta = dict(empresa=session.empresa, estado=tipo)
        if fechaOperacion:
            consulta['fechaOperacion'] = strftime(fechaOperacion)
        consulta['firma'] = cls._firma_consulta(consulta)
        try:
            resp = await session.post(endpoint, consulta)
            ordenes = pd.DataFrame.from_records(
                [item for item in resp['lst'] if item],  # type: ignore
            )
        except (NoOrdenesEncontradas, KeyError):
            # If resp does not have 'lst',
            # it means we do not have txns for that date
            ordenes = pd.DataFrame()
        else:
            ordenes.fechaOperacion = ordenes.fechaOperacion.apply(strptime)
            ordenes.estado = ordenes.estado.apply(lambda x: Estado(x))
        for col in (key for key in ordenes.keys() if key.startswith('ts')):
            ordenes[col] = ordenes[col].apply(cls._sanitize_timestamp)
        return ordenes

    @classmethod
    async def consulta_clave_rastreo_enviada(
        cls,
        claveRastreo: str,
        fechaOperacion: Optional[dt.date] = None,
    ):
        """
        La consulta por clave de rastreo solo está disponible
        para el día hábil en curso.
        https://stpmex.zendesk.com/hc/en-us/articles/360039782292-Query-Order-Sent-By-Tracking
        """
        endpoint = cls._endpoint + '/consOrdEnvRastreo'
        consulta = dict(
            empresa=session.empresa,
            claveRastreo=claveRastreo,
            institucionOperante=STP_BANK_CODE,
        )
        if fechaOperacion:
            consulta['fechaOperacion'] = strftime(fechaOperacion)
        consulta['firma'] = cls._firma_consulta(consulta)
        resp = await session.post(endpoint, consulta)
        return cls._sanitize_consulta(resp['ordenPago'])  # type: ignore

    @classmethod
    async def consulta_recibidas(  # type: ignore
        cls, fecha_operacion: Optional[dt.date] = None
    ) -> pd.DataFrame:
        """
        Consultar
        """
        return await cls._consulta_fecha(
            TipoOperacion.recibida, fecha_operacion
        )

    @classmethod
    async def consulta_enviadas(  # type: ignore
        cls, fecha_operacion: Optional[dt.date] = None
    ) -> pd.DataFrame:
        return await cls._consulta_fecha(
            TipoOperacion.enviada, fecha_operacion
        )

    @classmethod
    async def consulta(
        cls, fecha_operacion: Optional[dt.date] = None
    ) -> pd.DataFrame:
        recibidas, enviadas = await asyncio.gather(
            cls.consulta_recibidas(fecha_operacion),
            cls.consulta_enviadas(fecha_operacion),
        )
        return pd.concat([recibidas, enviadas])

    @staticmethod
    def _sanitize_timestamp(row):
        row /= 10**3  # convertir de milisegundos a segundos
        if row > 10**9:
            row = dt.datetime.fromtimestamp(row).strftime('%Y-%m-%d')
        return row



import datetime as dt
from unittest import mock

import pytest
from pydantic.dataclasses import dataclass

import aiostp
from aiostp.exc import NoPkeySet
from aiostp.http.client import Session
from aiostp.resources.base import Resource


@dataclass
class Mocker(Resource):
    name: str
    time: dt.date
    _endpoint = '123'
    _firma_fieldnames = ['1123']
    empresa = '123'


def test_to_dict():
    orden = aiostp.Orden(
        monto=29.00,
        conceptoPago='prueba',
        cuentaBeneficiario='072691004495711499',
        nombreBeneficiario='Pancho Villa',
        institucionContraparte=90646,
        cuentaOrdenante='646180157075068123',
    )

    orden_dict = orden.to_dict()

    assert orden_dict['monto'] == 29.00


def test_to_dict_w_date():
    mocker = Mocker(name='manu', time=dt.date(2020, 11, 29))

    mocker_dict = mocker.to_dict()

    assert mocker_dict


@pytest.mark.asyncio
async def test_no_pkey():
    session = Session()
    await session.configure(demo=True)

    with mock.patch('aiostp.resources.base.session', session):
        session.pkey = None
        base = Resource()
        with pytest.raises(NoPkeySet):
            base._firma_consulta(
                dict(
                    fechaOperacion='20201129',
                    claveRastreo='hola',
                    institucionOperante='STP',
                )
            )



from typing import List
from xml.etree import ElementTree

from pydantic.dataclasses import dataclass
from stpmex.auth import compute_signature
from stpmex.resources import Saldo as SaldoSTPMex

from ..exc import NoPkeySet
from ..http import session
from .base import Resource


@dataclass
class Saldo(Resource, SaldoSTPMex):
    @classmethod
    async def consulta_saldo_env_rec(cls) -> List['Saldo']:  # type: ignore
        data = dict(empresa=session.empresa, firma=cls._firma_consulta({}))
        resp = await session.post(cls._endpoint, data)
        saldos = []
        for saldo in resp['saldos']:  # type: ignore
            del saldo['empresa']
            saldos.append(cls(**saldo))  # type: ignore
        return saldos

    @classmethod
    async def consulta(cls) -> float:  # type: ignore
        """
        cuenta es la CLABE de la empresa

        Based on:
        https://stpmex.zendesk.com/hc/es/articles/360002812571-consultaSaldoCuenta
        """
        if not session.pkey:
            raise NoPkeySet()
        firma = compute_signature(session.pkey, session.clabe)
        data = f'''
<soapenv:Envelope
        xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/"
        xmlns:h2h="http://h2h.integration.spei.enlacefi.lgec.com/">
    <soapenv:Body>
        <h2h:consultaSaldoCuenta>
            <cuenta>{session.clabe}</cuenta>
            <firma>{firma}</firma>
        </h2h:consultaSaldoCuenta>
    </soapenv:Body>
</soapenv:Envelope>
'''
        resp = await session.soap_request('POST', '', data)
        root = ElementTree.fromstring(resp)
        saldo = root.findtext('.//saldo') or 0
        return float(saldo)
